@@ -19,6 +19,17 @@ database_engines = {}
 # logging.basicConfig(level=logging.DEBUG)
 
 
+def remove_nulls_from_dict_values(d):
+    for k in d:
+        if type(d[k]) == str:
+            d[k] = d[k].replace('\0', '')
+    return d
+
+
+def remove_nulls_from_dict_list(lo):
+    return list(map(remove_nulls_from_dict_values, lo))
+
+
 def get_table_key(table, obj=False):
     """This function takes an SQLAlchemy table object and returns its
     primary key, or if the table has no primary key, it returns its
@@ -664,10 +675,11 @@ class Reflection(models.Model):
         return source_table.c[source_key_name]
 
     def bulk_upsert(self, lst):
-        logging.debug(
+        logging.info(
             '{0} Saving {1} items'
             .format(self, len(lst))
         )
+        logging.info(lst)
         destination_db_engine = self.destination_database.mount()
         logging.debug('{0} retriving destination table'.format(self))
         destination_table = self.get_destination_table()
@@ -745,7 +757,7 @@ class Reflection(models.Model):
         for rec in targer_ids:
             if rec not in index_of_already_existing_records:
                 missing_records.append(rec)
-        logging.debug(
+        logging.info(
             '{0} {1} missing records were identified'
             .format(self, len(missing_records))
         )
@@ -756,7 +768,7 @@ class Reflection(models.Model):
             insert_data = [{destination_table_key.name: item}
                            for item in missing_records]
         if len(missing_records) > 0:
-            logging.debug(
+            logging.info(
                 '{0} Creating {1} missing records'
                 .format(self, len(missing_records))
             )
@@ -772,14 +784,14 @@ class Reflection(models.Model):
             )
         # run update statments for each record
         if len(lst) > 0:
-            logging.debug(
+            logging.info(
                 '{0} Updating data for {1} records'
                 .format(self, len(lst))
             )
             limit = 500
             start = 0
             while start < len(lst):
-                logging.debug(
+                logging.info(
                     '{0} Updating records from {1}, to {2}'
                     .format(self, start, start+limit)
                 )
@@ -787,16 +799,17 @@ class Reflection(models.Model):
                 with destination_db_engine.connect() as destination_dbc:
                     destination_dbc.execute(
                         text(self.reflection_statment),
-                        selected_items
+                        remove_nulls_from_dict_list(selected_items)
                     )
                 start += limit
         logging.debug('Done Saving'.format(self))
 
     def bulk_delete(self, lst):
-        logging.debug(
+        logging.info(
             '{0} Deleting {1} items'
             .format(self, len(lst))
         )
+        logging.info(lst)
         destination_table = self.get_destination_table()
         destination_table_key = get_table_key(destination_table, obj=True)
         targer_ids = lst
@@ -855,30 +868,37 @@ class Reflection(models.Model):
         )
         pdr_table = self.source_table.get_pdr_table()
         data_table = self.source_table.get_table()
-        data_table_key = get_table_key(data_table, obj=True)
-        upsert_stmt = pdr_table.select().with_only_columns([
+        pdr_stmt = pdr_table.select().with_only_columns([
             func.max(pdr_table.c.id).label('{0}_id'.format(pdr_prefix)),
-            func.max(pdr_table.c.c_action).label(
-                '{0}_c_action'.format(pdr_prefix)),
             pdr_table.c.c_record.label('{0}_c_record'.format(pdr_prefix)),
             func.max(pdr_table.c.c_time).label('{0}_c_time'.format(pdr_prefix))
         ])
         if self.last_commit:
-            upsert_stmt = upsert_stmt.where(pdr_table.c.id > self.last_commit)
-        upsert_stmt = upsert_stmt.group_by(pdr_table.c.c_record)
-        upsert_stmt = upsert_stmt.order_by('{0}_id'.format(pdr_prefix))
-        upsert_stmt = alias(upsert_stmt, 'pdr')
-        upsert_stmt = upsert_stmt.join(
+            pdr_stmt = pdr_stmt.where(pdr_table.c.id > self.last_commit)
+
+        pdr_stmt = pdr_stmt.group_by(pdr_table.c.c_record).order_by(
+            '{0}_id'.format(pdr_prefix))
+        subq_alias = 'pdr'
+        pdr_stmt = alias(pdr_stmt, subq_alias)
+        final_stmt = pdr_stmt.join(
             data_table,
-            upsert_stmt.c['{0}_c_record'.format(pdr_prefix)]
-            == self.get_source_key()
+            pdr_stmt.c['{0}_c_record'.format(pdr_prefix)]
+            == self.get_source_key(),
+            isouter=True
+        ).join(
+            pdr_table,
+            pdr_stmt.c['{0}_id'.format(pdr_prefix)] == pdr_table.c.id
         )
-        upsert_stmt = select([upsert_stmt])
+        final_stmt = select([final_stmt]).with_only_columns([
+            pdr_stmt,
+            pdr_table.c.c_action.label('{0}_c_action'.format(pdr_prefix)),
+            data_table
+        ], maintain_column_froms=True)
         with self.source_table.source_database.mount().connect() as source_dbc:
-            ret = source_dbc.execute(upsert_stmt)
+            ret = source_dbc.execute(final_stmt)
             commits = [dict(commit) for commit in ret.fetchall()]
 
-        def retrive_data_record(commit):
+        def remove_pdr_columns(commit):
             data_record = commit.copy()
             data_record.pop('{0}_id'.format(pdr_prefix))
             data_record.pop('{0}_c_action'.format(pdr_prefix))
@@ -887,7 +907,7 @@ class Reflection(models.Model):
             return data_record
 
         upserts = [
-            retrive_data_record(commit) for commit in commits
+            remove_pdr_columns(commit) for commit in commits
             if commit[
                 '{0}_c_action'.format(pdr_prefix)
             ] in ['INSERT', 'UPDATE']
@@ -900,7 +920,7 @@ class Reflection(models.Model):
             self.bulk_upsert(upserts)
         if len(deletes) > 0:
             if self.ignore_delete_events:
-                logging.debug(
+                logging.info(
                     '{0} Ignoring {1} delete events'
                     .format(self, len(deletes))
                 )
